# coding: utf-8
"""
    sickle.response
    ~~~~~~~~~~~~~~~

    :copyright: Copyright 2015 Mathias Loesch
"""

from lxml import etree

XMLParser = etree.XMLParser(remove_blank_text=True, recover=True, resolve_entities=False)


class OAIResponse(object):
    """A response from an OAI server.

    Provides access to the returned data on different abstraction
    levels.

    :param http_response: The original HTTP response.
    :param params: The OAI parameters for the request.
    :type params: dict
    """

    def __init__(self, http_response, params):
        self.params = params
        self.http_response = http_response

    @property
    def raw(self):
        """The server's response as unicode."""
        return self.http_response.text

    @property
    def xml(self):
        """The server's response as parsed XML."""
        return etree.XML(self.http_response.text.encode("utf8"),
                         parser=XMLParser)

    def __repr__(self):
        return '<OAIResponse %s>' % self.params.get('verb')



# coding: utf-8
"""
    sickle.iterator
    ~~~~~~~~~~~~~~~

    Collects classes for iterating over OAI responses

    :copyright: Copyright 2015 Mathias Loesch
"""

from sickle import oaiexceptions
from sickle.models import ResumptionToken


# Map OAI verbs to the XML elements
VERBS_ELEMENTS = {
    'GetRecord': 'record',
    'ListRecords': 'record',
    'ListIdentifiers': 'header',
    'ListSets': 'set',
    'ListMetadataFormats': 'metadataFormat',
    'Identify': 'Identify',
}


class BaseOAIIterator(object):
    """Iterator over OAI records/identifiers/sets transparently aggregated via
    OAI-PMH.

    Can be used to conveniently iterate through the records of a repository.

    :param sickle: The Sickle object that issued the first request.
    :type sickle: :class:`sickle.app.Sickle`
    :param params: The OAI arguments.
    :type params:  dict
    :param ignore_deleted: Flag for whether to ignore deleted records.
    :type ignore_deleted: bool
    """

    def __init__(self, sickle, params, ignore_deleted=False):
        self.sickle = sickle
        self.params = params
        self.ignore_deleted = ignore_deleted
        self.verb = self.params.get('verb')
        self.resumption_token = None
        self._next_response()

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.verb)

    def _get_resumption_token(self):
        """Extract and store the resumptionToken from the last response."""
        resumption_token_element = self.oai_response.xml.find(
            './/' + self.sickle.oai_namespace + 'resumptionToken')
        if resumption_token_element is None:
            return None
        token = resumption_token_element.text
        cursor = resumption_token_element.attrib.get('cursor', None)
        complete_list_size = resumption_token_element.attrib.get(
            'completeListSize', None)
        expiration_date = resumption_token_element.attrib.get(
            'expirationDate', None)
        resumption_token = ResumptionToken(
            token=token, cursor=cursor,
            complete_list_size=complete_list_size,
            expiration_date=expiration_date
        )
        return resumption_token

    def _next_response(self):
        """Get the next response from the OAI server."""
        params = self.params
        if self.resumption_token:
            params = {
                'resumptionToken': self.resumption_token.token,
                'verb': self.verb
            }
        self.oai_response = self.sickle.harvest(**params)
        error = self.oai_response.xml.find(
            './/' + self.sickle.oai_namespace + 'error')
        if error is not None:
            code = error.attrib.get('code', 'UNKNOWN')
            description = error.text or ''
            try:
                raise getattr(
                    oaiexceptions, code[0].upper() + code[1:])(description)
            except AttributeError:
                raise oaiexceptions.OAIError(description)
        self.resumption_token = self._get_resumption_token()

    def next(self):
        """Must be implemented by subclasses."""
        raise NotImplementedError


class OAIResponseIterator(BaseOAIIterator):
    """Iterator over OAI responses."""

    def next(self):
        """Return the next response."""
        while True:
            if self.oai_response:
                response = self.oai_response
                self.oai_response = None
                return response
            elif self.resumption_token:
                self._next_response()
            else:
                raise StopIteration


class OAIItemIterator(BaseOAIIterator):
    """Iterator over OAI records/identifiers/sets transparently aggregated via
    OAI-PMH.

    Can be used to conveniently iterate through the records of a repository.

    :param sickle: The Sickle object that issued the first request.
    :type sickle: :class:`sickle.app.Sickle`
    :param params: The OAI arguments.
    :type params:  dict
    :param ignore_deleted: Flag for whether to ignore deleted records.
    :type ignore_deleted: bool
    """

    def __init__(self, sickle, params, ignore_deleted=False):
        self.mapper = sickle.class_mapping[params.get('verb')]
        self.element = VERBS_ELEMENTS[params.get('verb')]
        super(OAIItemIterator, self).__init__(sickle, params, ignore_deleted)

    def _next_response(self):
        super(OAIItemIterator, self)._next_response()
        self._items = self.oai_response.xml.iterfind(
            './/' + self.sickle.oai_namespace + self.element)

    def next(self):
        """Return the next record/header/set."""
        while True:
            for item in self._items:
                mapped = self.mapper(item)
                if self.ignore_deleted and mapped.deleted:
                    continue
                return mapped
            if self.resumption_token and self.resumption_token.token:
                self._next_response()
            else:
                raise StopIteration



# coding: utf-8
"""
    sickle
    ~~~~~~

    OAI-PMH for Humans.

    :copyright: Copyright 2015 Mathias Loesch
"""

__version__ = '0.6.2'

from .app import Sickle
from sickle.response import OAIResponse



# coding: utf-8
"""
    sickle.app
    ~~~~~~~~~~

    An OAI-PMH client.

    :copyright: Copyright 2015 Mathias Loesch
"""
import inspect
import logging
import time

import requests
from sickle.iterator import BaseOAIIterator, OAIItemIterator
from sickle.response import OAIResponse

from .models import (Set, Record, Header, MetadataFormat,
                     Identify)

logger = logging.getLogger(__name__)

OAI_NAMESPACE = '{http://www.openarchives.org/OAI/%s/}'


# Map OAI verbs to class representations
DEFAULT_CLASS_MAP = {
    'GetRecord': Record,
    'ListRecords': Record,
    'ListIdentifiers': Header,
    'ListSets': Set,
    'ListMetadataFormats': MetadataFormat,
    'Identify': Identify,
}


class Sickle(object):
    """Client for harvesting OAI interfaces.

    Use it like this::

        >>> sickle = Sickle('http://elis.da.ulcc.ac.uk/cgi/oai2')
        >>> records = sickle.ListRecords(metadataPrefix='oai_dc')
        >>> records.next()
        <Record oai:eprints.rclis.org:3780>

    :param endpoint: The endpoint of the OAI interface.
    :type endpoint: str
    :param http_method: Method used for requests (GET or POST, default: GET).
    :type http_method: str
    :param protocol_version: The OAI protocol version.
    :type protocol_version: str
    :param iterator: The type of the returned iterator
           (default: :class:`sickle.iterator.OAIItemIterator`)
    :param max_retries: Number of retries if HTTP request fails.
    :type max_retries: int
    :type protocol_version: str
    :param class_mapping: A dictionary that maps OAI verbs to classes representing
                          OAI items. If not provided,
                          :data:`sickle.app.DEFAULT_CLASS_MAPPING` will be used.
    :type class_mapping: dict
    :param encoding:     Can be used to override the encoding used when decoding
                         the server response. If not specified, `requests` will
                         use the encoding returned by the server in the
                         `content-type` header. However, if the `charset`
                         information is missing, `requests` will fallback to
                         `'ISO-8859-1'`.
    :type encoding:      str
    :param request_args: Arguments to be passed to requests when issuing HTTP
                         requests. Useful examples are `auth=('username', 'password')`
                         for basic auth-protected endpoints or `timeout=<int>`.
                         See the `documentation of requests <http://docs.python-requests.org/en/master/api/#main-interface>`_
                         for all available parameters.
    """

    def __init__(self, endpoint, http_method='GET', protocol_version='2.0',
                 iterator=OAIItemIterator, max_retries=5,
                 class_mapping=None, encoding=None, **request_args):
        self.endpoint = endpoint
        if http_method not in ['GET', 'POST']:
            raise ValueError("Invalid HTTP method: %s! Must be GET or POST.")
        if protocol_version not in ['2.0', '1.0']:
            raise ValueError(
                "Invalid protocol version: %s! Must be 1.0 or 2.0.")
        self.http_method = http_method
        self.protocol_version = protocol_version
        if inspect.isclass(iterator) and issubclass(iterator, BaseOAIIterator):
            self.iterator = iterator
        else:
            raise TypeError(
                "Argument 'iterator' must be subclass of %s" % BaseOAIIterator.__name__)
        self.max_retries = max_retries
        self.oai_namespace = OAI_NAMESPACE % self.protocol_version
        self.class_mapping = class_mapping or DEFAULT_CLASS_MAP
        self.encoding = encoding
        self.request_args = request_args

    def harvest(self, **kwargs):  # pragma: no cover
        """Make HTTP requests to the OAI server.

        :param kwargs: OAI HTTP parameters.
        :rtype: :class:`sickle.OAIResponse`
        """
        for _ in range(self.max_retries):
            if self.http_method == 'GET':
                http_response = requests.get(self.endpoint, params=kwargs,
                                             **self.request_args)
            else:
                http_response = requests.post(self.endpoint, data=kwargs,
                                              **self.request_args)
            if http_response.status_code == 503:
                try:
                    retry_after = int(http_response.headers.get('retry-after'))
                except TypeError:
                    retry_after = 20
                logger.info(
                    "HTTP 503! Retrying after %d seconds..." % retry_after)
                time.sleep(retry_after)
            else:
                http_response.raise_for_status()
                if self.encoding:
                    http_response.encoding = self.encoding
                return OAIResponse(http_response, params=kwargs)

    def ListRecords(self, ignore_deleted=False, **kwargs):
        """Issue a ListRecords request.

        :param ignore_deleted: If set to :obj:`True`, the resulting
                              iterator will skip records flagged as deleted.
        :rtype: :class:`sickle.iterator.BaseOAIIterator`
        """
        params = kwargs
        params.update({'verb': 'ListRecords'})
        # noinspection PyCallingNonCallable
        return self.iterator(self, params, ignore_deleted=ignore_deleted)

    def ListIdentifiers(self, ignore_deleted=False, **kwargs):
        """Issue a ListIdentifiers request.

        :param ignore_deleted: If set to :obj:`True`, the resulting
                              iterator will skip records flagged as deleted.
        :rtype: :class:`sickle.iterator.BaseOAIIterator`
        """
        params = kwargs
        params.update({'verb': 'ListIdentifiers'})
        return self.iterator(self,
                             params, ignore_deleted=ignore_deleted)

    def ListSets(self, **kwargs):
        """Issue a ListSets request.

        :rtype: :class:`sickle.iterator.BaseOAIIterator`
        """
        params = kwargs
        params.update({'verb': 'ListSets'})
        return self.iterator(self, params)

    def Identify(self):
        """Issue an Identify request.

        :rtype: :class:`sickle.models.Identify`
        """
        params = {'verb': 'Identify'}
        return Identify(self.harvest(**params))

    def GetRecord(self, **kwargs):
        """Issue a ListSets request."""
        params = kwargs
        params.update({'verb': 'GetRecord'})
        record = self.iterator(self, params).next()
        return record

    def ListMetadataFormats(self, **kwargs):
        """Issue a ListMetadataFormats request.

        :rtype: :class:`sickle.iterator.BaseOAIIterator`
        """
        params = kwargs
        params.update({'verb': 'ListMetadataFormats'})
        return self.iterator(self, params)



# coding: utf-8
"""
    sickle.utils
    ~~~~~~~~~~~~

    Collects utility functions.

    :copyright: Copyright 2015 Mathias Loesch
"""

import re
from collections import defaultdict


def get_namespace(element):
    """Return the namespace of an XML element.

    :param element: An XML element.
    """
    return re.search('(\{.*\})', element.tag).group(1)


def xml_to_dict(tree, paths=None, nsmap=None, strip_ns=False):
    """Convert an XML tree to a dictionary.

    :param tree: etree Element
    :type tree: :class:`lxml.etree._Element`
    :param paths: An optional list of XPath expressions applied on the XML tree.
    :type paths: list[basestring]
    :param nsmap: An optional prefix-namespace mapping for conciser spec of paths.
    :type nsmap: dict
    :param strip_ns: Flag for whether to remove the namespaces from the tags.
    :type strip_ns: bool
    """
    paths = paths or ['.//']
    nsmap = nsmap or {}
    fields = defaultdict(list)
    for path in paths:
        elements = tree.findall(path, nsmap)
        for element in elements:
            tag = re.sub(
                r'\{.*\}', '', element.tag) if strip_ns else element.tag
            fields[tag].append(element.text)
    return dict(fields)



# coding: utf-8
"""
    sickle.models
    ~~~~~~~~~~~~~

    Collects classes for OAI-specific entities.

    :copyright: Copyright 2015 Mathias Loesch
"""

from lxml import etree

from .utils import get_namespace, xml_to_dict
from ._compat import PY3


class ResumptionToken(object):
    """Represents a resumption token."""

    def __init__(self, token='', cursor='', complete_list_size='',
                 expiration_date=''):
        self.token = token
        self.cursor = cursor
        self.complete_list_size = complete_list_size
        self.expiration_date = expiration_date

    def __repr__(self):
        return '<ResumptionToken %s>' % self.token


class OAIItem(object):
    """A generic OAI item.

    :param xml: XML representation of the entity.
    :param strip_ns: Flag for whether to remove the namespaces from the
                     element names in the dictionary representation.
    """

    def __init__(self, xml, strip_ns=True):
        super(OAIItem, self).__init__()

        #: The original parsed XML
        self.xml = xml
        self._strip_ns = strip_ns
        self._oai_namespace = get_namespace(self.xml)


    def __bytes__(self):
        return etree.tounicode(self.xml).encode("utf8")

    def __str__(self):
        return self.__unicode__() if PY3 else self.__bytes__()

    def __unicode__(self):
        return etree.tounicode(self.xml)

    @property
    def raw(self):
        """The original XML as unicode."""
        return etree.tounicode(self.xml)


class Identify(OAIItem):
    """Represents an Identify container.

    This object differs from the other entities in that is has to be created
    from a :class:`sickle.response.OAIResponse` instead of an XML element.

    :param identify_response: The response for an Identify request.
    :type identify_response: :class:`sickle.OAIResponse`
    """

    def __init__(self, identify_response):
        super(Identify, self).__init__(identify_response.xml, strip_ns=True)
        self.xml = self.xml.find('.//' + self._oai_namespace + 'Identify')
        self._identify_dict = xml_to_dict(self.xml, strip_ns=True)
        for k, v in self._identify_dict.items():
            setattr(self, k.replace('-', '_'), v[0])

    def __repr__(self):
        return '<Identify>'

    def __iter__(self):
        return iter(self._identify_dict.items()) if PY3 else \
            self._identify_dict.iteritems()


class Header(OAIItem):
    """Represents an OAI Header.

    :param header_element: The XML element 'header'.
    :type header_element: :class:`lxml.etree._Element`
    """

    def __init__(self, header_element):
        super(Header, self).__init__(header_element, strip_ns=True)
        self.deleted = self.xml.attrib.get('status') == 'deleted'
        _identifier_element = self.xml.find(self._oai_namespace + 'identifier')
        _datestamp_element = self.xml.find(self._oai_namespace + 'datestamp')

        self.identifier = getattr(_identifier_element, 'text', None)
        self.datestamp = getattr(_datestamp_element, 'text', None)
        self.setSpecs = [setSpec.text for setSpec in
                         self.xml.findall(self._oai_namespace + 'setSpec')]

    def __repr__(self):
        if self.deleted:
            return '<Header %s [deleted]>' % self.identifier
        else:
            return '<Header %s>' % self.identifier

    def __iter__(self):
        return iter([
            ('identifier', self.identifier),
            ('datestamp', self.datestamp),
            ('setSpecs', self.setSpecs)
        ])


class Record(OAIItem):
    """Represents an OAI record.

    :param record_element: The XML element 'record'.
    :type record_element: :class:`lxml.etree._Element`
    :param strip_ns: Flag for whether to remove the namespaces from the
                     element names.
    """

    def __init__(self, record_element, strip_ns=True):
        super(Record, self).__init__(record_element, strip_ns=strip_ns)
        self.header = Header(self.xml.find(
            './/' + self._oai_namespace + 'header'))
        self.deleted = self.header.deleted
        if not self.deleted:
            # We want to get record/metadata/<container>/*
            # <container> would be the element ``dc``
            # in the ``oai_dc`` case.
            self.metadata = xml_to_dict(
                self.xml.find(
                    './/' + self._oai_namespace + 'metadata'
                ).getchildren()[0], strip_ns=self._strip_ns)

    def __repr__(self):
        if self.header.deleted:
            return '<Record %s [deleted]>' % self.header.identifier
        else:
            return '<Record %s>' % self.header.identifier

    def __iter__(self):
        return iter(self.metadata.items()) if PY3 else \
            self.metadata.iteritems()


class Set(OAIItem):
    """Represents an OAI set.

    :param set_element: The XML element 'set'.
    :type set_element: :class:`lxml.etree._Element`
    """

    def __init__(self, set_element):
        super(Set, self).__init__(set_element, strip_ns=True)
        self._set_dict = xml_to_dict(self.xml, strip_ns=True)
        for k, v in self._set_dict.items():
            setattr(self, k.replace('-', '_'), v[0])

    def __repr__(self):
        return u'<Set %s>'.encode('utf8') % self.setName

    def __iter__(self):
        return iter(self._set_dict.items()) if PY3 else \
            self._set_dict.iteritems()


class MetadataFormat(OAIItem):
    """Represents an OAI MetadataFormat.

    :param mdf_element: The XML element 'metadataFormat'.
    :type mdf_element: :class:`lxml.etree._Element`
    """

    def __init__(self, mdf_element):
        super(MetadataFormat, self).__init__(mdf_element, strip_ns=True)
        #: The prefix of this format.
        self._mdf_dict = xml_to_dict(self.xml, strip_ns=True)
        for k, v in self._mdf_dict.items():
            setattr(self, k.replace('-', '_'), v[0])

    def __repr__(self):
        return u'<MetadataFormat %s>'.encode('utf8') % self.metadataPrefix

    def __iter__(self):
        return iter(self._mdf_dict.items()) if PY3 else \
            self._mdf_dict.iteritems()



# coding: utf-8
"""
    sickle._compat
    ~~~~~~~~~~~~~~

    Python 2/3 compatibility layer

    :copyright: Copyright 2015 Mathias Loesch
"""

import sys

PY3 = sys.version_info >= (3, 0)

if PY3:  # pragma: no cover
    string_types = str,
    text_type = str
    binary_type = bytes
else:  # pragma: no cover
    string_types = basestring,
    text_type = unicode
    binary_type = str


def to_unicode(x):  # pragma: no cover
    """Convert argument into a unicode string."""
    if PY3:
        return str(x)
    else:
        return x.decode('utf-8')



# coding: utf-8

from setuptools import setup

setup(
    name='Sickle',
    version='0.6.2',
    url='http://github.com/mloesch/sickle',
    license='BSD',
    author='Mathias Loesch',
    description='A lightweight OAI client library for Python',
    long_description=open('README.rst').read() + '\n\n' +
    open('CHANGES.rst').read(),
    packages=['sickle'],
    platforms='any',
    tests_require=[
        'nose>=1.0',
        'mock>=1.0.1'
    ],
    install_requires=[
        'requests>=1.1.0',
        'lxml>=3.2.3'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Text Processing :: Markup :: XML',
    ],
    test_suite="sickle.tests",
    keywords="oai oai-pmh",
    zip_safe=False,
)



# coding: utf-8
"""
    sickle.oaiexceptions
    ~~~~~~~~~~~~~~~~~~~~

    OAI errors.

    :copyright: Copyright 2015 Mathias Loesch
"""


class BadArgument(Exception):
    """
    The request includes illegal arguments, is missing required arguments,
    includes a repeated argument, or values for arguments have an illegal
    syntax.
    """
    pass


class BadVerb(Exception):
    """
    Value of the verb argument is not a legal OAI-PMH verb, the verb argument
    is missing, or the verb argument is repeated.
    """
    pass


class BadResumptionToken(Exception):
    """
    The value of the resumptionToken argument is invalid or expired.
    """
    pass


class CannotDisseminateFormat(Exception):
    """
    The metadata format identified by the value given for the metadataPrefix
    argument is not supported by the item or by the repository.
    """
    pass


class IdDoesNotExist(Exception):
    """
    The value of the identifier argument is unknown or illegal in this
    repository.
    """
    pass


class NoSetHierarchy(Exception):
    """
    The repository does not support sets.
    """
    pass


class NoMetadataFormat(Exception):
    """
    There are no metadata formats available for the specified item.
    """
    pass


class NoRecordsMatch(Exception):
    """
    The combination of the values of the from, until, set and metadataPrefix
    arguments results in an empty list.
    """
    pass


class OAIError(Exception):
    """
    Context specific OAI errors not covered by the classes above.
    """
    pass
